/*
 * Driver for Data Translation DT2811
 * David A. Schleef
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <asm/io.h>
#include <dac.h>

/*

0x00    ADCSR R/W  A/D Control/Status Register
  bit 7 - (R) 1 indicates A/D conversion done
          reading ADDAT clears bit
          (W) ignored
  bit 6 - (R) 1 indicates A/D error
          (W) ignored
  bit 5 - (R) 1 indicates A/D busy, cleared at end
          of conversion
          (W) ignored
  bit 4 - (R) 0
          (W) 
  bit 3 - (R) 0
  bit 2 - (R/W) 1 indicates interrupts enabled
  bits 1,0 - (R/W) mode bits
     00  single conversion on ADGCR load
     01  continuous conversion, internal clock,
         (clock enabled on ADGCR load)
     10  continuous conversion, internal clock,
         external trigger
     11  continuous conversion, external clock,
         external trigger

0x01    ADGCR R/W A/D Gain/Channel Register
  bit 6,7 - (R/W) gain select
    00  gain=1, both PGH, PGL models
    01  gain=2 PGH, 10 PGL
    10  gain=4 PGH, 100 PGL
    11  gain=8 PGH, 500 PGL
  bit 4,5 - reserved
  bit 3-0 - (R/W) channel select
    channel number from 0-15

0x02,0x03 (R) ADDAT A/D Data Register
          (W) DADAT0 D/A Data Register 0
  0x02 low byte
  0x03 high byte

0x04,0x05 (W) DADAT0 D/A Data Register 1

0x06 (R) DIO0 Digital Input Port 0
     (W) DIO1 Digital Output Port 1

0x07 TMRCTR (R/W) Timer/Counter Register
  bits 6,7 - reserved
  bits 5-3 - Timer frequency control (mantissa)
    543  divisor  freqency (kHz)
    000  1        600
    001  10       60
    010  2        300
    011  3        200
    100  4        150
    101  5        120
    110  6        100
    111  12       50
  bits 2-0 - Timer frequency control (exponent)
    543  multiply divisor/divide frequency by
    000  1
    001  10
    010  100
    011  1000
    100  10000
    101  100000
    110  1000000
    111  10000000

*/

#define DT2811_SIZE 8

#define DT2811_ADCSR 0
#define DT2811_ADGCR 1
#define DT2811_ADDATLO 2
#define DT2811_ADDATHI 3
#define DT2811_DADAT0LO 2
#define DT2811_DADAT0HI 3
#define DT2811_DADAT1LO 4
#define DT2811_DADAT1HI 5
#define DT2811_DIO 6
#define DT2811_TMRCTR 7

/*
 * flags
 */

/* ADCSR */

#define DT2811_ADDONE   0x80
#define DT2811_ADERROR  0x40
#define DT2811_ADBUSY   0x20
#define DT2811_CLRERROR 0x10
#define DT2811_INTENB   0x04
#define DT2811_ADMODE   0x03

dac_devinit dt2811_init;
int dt2811_rem(dac_device *dev);
int dt2811_ad(dac_device *,dac_s *);
int dt2811_da(dac_device *,dac_s *);
int dt2811_di(dac_device *,dac_s *);
int dt2811_do(dac_device *,dac_s *);

#if 0
static void dt2811_interrupt(int irq,void *dev,struct pt_regs * regs)
{
	int lo,hi,usec;
	dac_device *d=dev;
	struct timeval it;
	dac_sample pt;


switch(d->i_admode){
case DAC_MDEMAND:
	lo=inb(d->iobase+DT2811_ADDATLO);
	hi=inb(d->iobase+DT2811_ADDATHI);

	pt.data=lo+0x100*hi;
	pt.chan=d->curadchan;
	pt.time=trigtime.tv_usec;
	copy_to_buffer(&(d->rdbuf),&pt,sizeof(dac_sample));

	if(d->ntrig>0){
		d->ntrig--;
		outb(d->curadchan,d->iobase+DT2811_ADGCR);
		do_gettimeofday(&trigtime);
	}else{
		wake_up_interruptible(&d->rdwait);
	}
	break;
case DAC_MCONTS:
	lo=inb(d->iobase+DT2811_ADDATLO);
	hi=inb(d->iobase+DT2811_ADDATHI);
	do_gettimeofday(&it);
	if(d->ntrig>0){
		/*do_gettimeofday(&it);*/
		d->ntrig--;
		/*pt.data=lo+0x100*hi;*/
		pt.data=it.tv_usec;
		pt.chan=d->curadchan;
		pt.time=trigtime.tv_usec;
		copy_to_buffer(&(d->rdbuf),&pt,sizeof(dac_sample));
		wake_up_interruptible(&d->rdwait);
	}
	if(inb(d->iobase+DT2811_ADCSR)&DT2811_ADERROR){
		printk("dt2811: ad error %d %d\n",it.tv_sec,it.tv_usec);
		outb(DT2811_CLRERROR|DT2811_INTENB|1,d->iobase+DT2811_ADCSR);
	}
	trigtime.tv_usec+=10000;
	if(trigtime.tv_usec>=1000000){
		trigtime.tv_usec-=1000000;
		trigtime.tv_sec++;
	}
	break;
}
}
#endif

int dt2811_init(dac_device *dev,dac_devconfig *it)
{
	int i,irqs,irq;
	long flags;

	if(strcmp("dt2811",it->name))
		return 0;

	printk("dac%d: dt2811: base=0x%04x ",dev->minor,it->iobase);
	dev->iobase=it->iobase;
	dev->iosize=DT2811_SIZE;
	if(check_region(dev->iobase,DT2811_SIZE)<0){
		printk("I/O port conflict\n");
		return -EIO;
	}
	request_region(dev->iobase,DT2811_SIZE,"dt2811");

	outb(0,dev->iobase+DT2811_ADCSR);
	udelay(100);
	i=inb(dev->iobase+DT2811_ADDATLO);
	i=inb(dev->iobase+DT2811_ADDATHI);

#ifdef INTERRUPTS
	save_flags(flags);
	sti();
	irqs=probe_irq_on();

	outb(DT2811_CLRERROR|DT2811_INTENB,dev->iobase+DT2811_ADCSR);
	outb(0,dev->iobase+DT2811_ADGCR);

	udelay(100);

	irq=probe_irq_off(irqs);
	restore_flags(flags);

	/*outb(DT2811_CLRERROR|DT2811_INTENB,dev->iobase+DT2811_ADCSR);*/
#endif

	irq=0;
	dev->irq=0;
	if(irq>0){
		i=inb(dev->iobase+DT2811_ADDATLO);
		i=inb(dev->iobase+DT2811_ADDATHI);
		printk("(irq = %d)\n",irq);
		/* request_irq(irq,dt2811_interrupt,0,"dt2811",dev); */
		dev->irq=irq;
	}else if(irq==0){
		printk("(no irq)\n");
	}else{
		printk("( multiple irq's -- this is bad! )\n");
	}

	dev->chan=20;
	dev->chinfo=kmalloc(dev->chan*sizeof(dac_chinfo),GFP_KERNEL);
	if(!dev->chinfo)return -ENOMEM;

	for(i=0;i<16;i++){
		dev->chinfo[i].flags=DAC_ANALOG|DAC_INPUT;
		dev->chinfo[i].trig=dt2811_ad;
		dev->chinfo[i].bits=12;
		dev->chinfo[i].zero=-5.0;
		dev->chinfo[i].delta=10.0/4096;
	}
	
	dev->chinfo[16].flags=DAC_ANALOG|DAC_OUTPUT;
	dev->chinfo[16].trig=dt2811_da;
	dev->chinfo[16].bits=12;
	dev->chinfo[16].zero=-5.0;
	dev->chinfo[16].delta=10.0/4096;
	
	dev->chinfo[17].flags=DAC_ANALOG|DAC_OUTPUT;
	dev->chinfo[17].trig=dt2811_da;
	dev->chinfo[17].bits=12;
	dev->chinfo[17].zero=-5.0;
	dev->chinfo[17].delta=10.0/4096;
	
	dev->chinfo[18].flags=DAC_DIGITAL|DAC_INPUT;
	dev->chinfo[18].trig=dt2811_di;
	dev->chinfo[18].bits=8;
	dev->chinfo[18].zero=0;
	dev->chinfo[18].delta=0;
	
	dev->chinfo[19].flags=DAC_DIGITAL|DAC_OUTPUT;
	dev->chinfo[19].trig=dt2811_do;
	dev->chinfo[19].bits=8;
	dev->chinfo[19].zero=0;
	dev->chinfo[19].delta=0;

	dac_initbuf(&dev->rdbuf);
	
	dev->rem=dt2811_rem;

	return 1;
}


int dt2811_rem(dac_device *dev)
{
	printk("dac%d: dt2811: remove\n",dev->minor);
	kfree(dev->chinfo);

	if(dev->irq){
		free_irq(dev->irq,dev);
	}
	release_region(dev->iobase,dev->iosize);

	dac_freebuf(&dev->rdbuf);
	
	return 0;
}

int dt2811_ad(dac_device *dev,dac_s *it)
{
	int lo,hi;
	
	outb(it->chan,dev->iobase+DT2811_ADGCR);
#ifdef TRIGTIME
	do_gettimeofday(&it->trigtime);
#endif
	while(inb(dev->iobase+DT2811_ADCSR)&DT2811_ADBUSY)
		udelay(25);
	lo=inb(dev->iobase+DT2811_ADDATLO);
	hi=inb(dev->iobase+DT2811_ADDATHI);

	it->data=lo+0x100*hi;
	
	return 0;
}

#if 0
int dt2811_adtrig(kdev_t minor,dac_adtrig *adtrig)
{
	dac_device *dev=dac_devices+minor;

	if(adtrig->n<1)return 0;
	dev->curadchan=adtrig->chan;
switch(dev->i_admode){
case DAC_MDEMAND:
	dev->ntrig=adtrig->n-1;
	/*printk("dt2811: AD soft trigger\n");*/
	/*outb(DT2811_CLRERROR|DT2811_INTENB,dev->iobase+DT2811_ADCSR);*/  /* not neccessary */
	outb(dev->curadchan,dev->iobase+DT2811_ADGCR);
	do_gettimeofday(&trigtime);
	break;
case DAC_MCONTS:
	dev->ntrig=adtrig->n;
	break;
}

	return 0;
}
#endif

int dt2811_da(dac_device *dev,dac_s *it)
{
	int lo,hi;

	lo=(it->data&0xff);
	hi=(it->data>>8)&0x0f;
	
	switch(it->chan){
	case 16:
		outb(lo,dev->iobase+DT2811_DADAT0LO);
		outb(hi,dev->iobase+DT2811_DADAT0HI);
		break;
	case 17:
		outb(lo,dev->iobase+DT2811_DADAT1LO);
		outb(hi,dev->iobase+DT2811_DADAT1HI);
		break;
	}
#ifdef TRIGTIME
	do_gettimeofday(&it->trigtime);
#endif
	return 0;
}

int dt2811_di(dac_device *dev,dac_s *it)
{
	it->data=inb(it->data,dev->iobase+DT2811_DIO);
#ifdef TRIGTIME
	do_gettimeofday(&it->trigtime);
#endif

	return 0;
}

int dt2811_do(dac_device *dev,dac_s *it)
{
	outb(it->data,dev->iobase+DT2811_DIO);
#ifdef TRIGTIME
	do_gettimeofday(&it->trigtime);
#endif

	return 0;
}

