/*
 * Driver for Data Translation DT2814
 * David A. Schleef
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <dac.h>



#define DT2814_SIZE 2

#define DT2814_CSR 0
#define DT2814_DATA 1

/*
 * flags
 */

#define DT2814_FINISH 0x80
#define DT2814_ERR 0x40
#define DT2814_BUSY 0x20
#define DT2814_ENB 0x10
#define DT2814_CHANMASK 0x0f

dac_devinit dt2814_init;
int dt2814_ad(dac_device *dev,dac_s *ad);
int dt2814_rem(dac_device *dev);


int dt2814_ad(dac_device *dev,dac_s *ad)
{
	int hi,lo;

	outb(ad->chan,dev->iobase+DT2814_CSR);
#ifdef TRIGTIME
	do_gettimeofday(&ad->trigtime);
#endif
	while(!(inb(dev->iobase+DT2814_CSR)&DT2814_FINISH)){
		udelay(25);
	}
	hi=inb(dev->iobase+DT2814_DATA);
	lo=inb(dev->iobase+DT2814_DATA);

	ad->data=(hi<<4)|(lo>>4);
	
	return 0;
}

int dt2814_init(dac_device *dev,dac_devconfig *it)
{
	int i,irqs,irq;
	long flags;

	if(strcmp("dt2814",it->name))
		return 0;

	printk("dac%d: dt2814: 0x%04x ",dev->minor,it->iobase);
	if(check_region(it->iobase,DT2814_SIZE)<0){
		printk("I/O port conflict\n");
		return -EIO;
	}
	request_region(it->iobase,DT2814_SIZE,"dt2814");
	dev->iobase=it->iobase;
	dev->iosize=DT2814_SIZE;

	outb(0,dev->iobase+DT2814_CSR);
	udelay(100);
	if(inb(dev->iobase+DT2814_CSR)&DT2814_ERR){
		printk("reset error (fatal)\n");
		return -EIO;
	}
	i=inb(dev->iobase+DT2814_DATA);
	i=inb(dev->iobase+DT2814_DATA);

	irq=it->irq;
	if(irq<0){
#ifdef INTERRUPTS
		save_flags(flags);
		sti();
		irqs=probe_irq_on();
	
		outb(0,dev->iobase+DT2814_CSR);

		udelay(100);

		irq=probe_irq_off(irqs);
		restore_flags(flags);
		if(inb(dev->iobase+DT2814_CSR)&DT2814_ERR){
			printk("error probing irq (bad) \n");
		}

		i=inb(dev->iobase+DT2814_DATA);
		i=inb(dev->iobase+DT2814_DATA);
#else /* !INTERRUPTS */
		irq=0;
#endif
	}
	dev->irq=0;
	if(irq>0){
		printk("( irq = %d )\n",irq);
/*
		request_irq(irq,dt2814_interrupt,SA_INTERRUPT,"dt2814",dev);
*/
		dev->irq=irq;
	}else if(irq==0){
		printk("(no irq)\n");
	}else{
		printk("(probe returned multiple irqs--bad)\n");
	}
	
	dev->chan=16;

	dev->chinfo=kmalloc(dev->chan*sizeof(dac_chinfo),GFP_KERNEL);
	if(!dev->chinfo){
		return -ENOMEM;
	}
	for(i=0;i<16;i++){
		dev->chinfo[i].flags=DAC_ANALOG|DAC_INPUT;
		dev->chinfo[i].trig=dt2814_ad;
		dev->chinfo[i].bits=12;
		dev->chinfo[i].zero=-5.0;
		dev->chinfo[i].delta=10.0/4096;
	}

	dac_initbuf(&dev->rdbuf);

	dev->rem=dt2814_rem;

	return 1;
}


int dt2814_rem(dac_device *dev)
{
	printk("dac%d: dt2814: remove\n",dev->minor);
	
	kfree(dev->chinfo);

	if(dev->irq){
		free_irq(dev->irq,dev);
	}
	release_region(dev->iobase,dev->iosize);

	dac_freebuf(&dev->rdbuf);

	return 0;
}


