/*
 * Driver for Data Translation DT2821 series
 * David A. Schleef
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <dac.h>



#define DT2821_SIZE 0x10

#define DT2821_ADCSR 0
#define DT2821_CHANCSR 2
#define DT2821_ADDAT 4
#define DT2821_DACSR 6
#define DT2821_DADAT 8
#define DT2821_DIODAT 10
#define DT2821_SUPCSR 12
#define DT2821_TMRCTR 14

/*
 * flags
 */

/* ADCSR */

#define DT2821_ADCSR_MASK 0xfff0
#define DT2821_ADCSR_VAL 0x7c00

#define DT2821_ADERR 0x8000
#define DT2821_ADCLK 0x0200
#define DT2821_MUXBUSY 0x0100
#define DT2821_ADDONE 0x0080
#define DT2821_IADDONE 0x0040

/* CHANCSR */

#define DT2821_CHANCSR_MASK 0xfff0
#define DT2821_CHANCSR_VAL 0x70f0

#define DT2821_LLE 0x8000

/* DACSR */

#define DT2821_DACSR_MASK 0xffff
#define DT2821_DACSR_VAL 0x7c9c

#define DT2821_DAERR 0x8000
#define DT2821_YSEL 0x0200
#define DT2821_SSEL 0x0100
#define DT2821_DACRDY 0x0080
#define DT2821_IDARDY 0x0040
#define DT2821_DACLK 0x0020
#define DT2821_HBOE 0x0002
#define DT2821_LBOE 0x0001

/* SUPCSR */

#define DT2821_SUPCSR_MASK 0xfcff
#define DT2821_SUPCSR_VAL 0x0000

#define DT2821_DMAD 0x8000
#define DT2821_ERRINTEN 0x4000
#define DT2821_CLRDMADNE 0x2000
#define DT2821_DDMA 0x1000
#define DT2821_BUFFB 0x0200
#define DT2821_SCDN 0x0100
#define DT2821_DACON 0x0080
#define DT2821_ADCINIT 0x0040
#define DT2821_DACINIT 0x0020
#define DT2821_PRLD 0x0010
#define DT2821_STRIG 0x0008
#define DT2821_XTRIG 0x0004
#define DT2821_XCLK 0x0002
#define DT2821_BDINIT 0x0001

/* TMRCTR */

#define DT2821_TMRCTR_MASK 0xff00
#define DT2821_TMRCTR_VAL 0xf000


dac_devinit dt282x_init;
int dt282x_rem(dac_device *dev);

int dt282x_ad(dac_device *dev,dac_s *it)
{
	outw(DT2821_LLE,dev->iobase+DT2821_CHANCSR);
	outw(DT2821_ADCLK|(0<<4)|(it->chan),dev->iobase+DT2821_ADCSR);
	outw(0,dev->iobase+DT2821_CHANCSR);
	outw(DT2821_PRLD,dev->iobase+DT2821_SUPCSR);
	while(inw(dev->iobase+DT2821_ADCSR)&DT2821_MUXBUSY)
		udelay(5);
	outw(DT2821_STRIG,dev->iobase+DT2821_SUPCSR);
	while(!(inw(dev->iobase+DT2821_ADCSR)&DT2821_ADDONE))
		udelay(5);
	it->data=(short)inw(dev->iobase+DT2821_ADDAT);

	return 0;
}

int dt282x_da(dac_device *dev,dac_s *it)
{
	outw(DT2821_SSEL|((it->chan==17)?DT2821_YSEL:0),
		dev->iobase+DT2821_DACSR);
	outw(it->data,dev->iobase+DT2821_DADAT);
	outw(DT2821_DACON,dev->iobase+DT2821_SUPCSR);

	return 0;
}

int dt282x_di(dac_device *dev,dac_s *it)
{
	return 0;
}

int dt282x_do(dac_device *dev,dac_s *it)
{
	return 0;
}

int dt282x_init(dac_device *dev,dac_devconfig *it)
{
	int i,irqs,irq;
	long flags;

	if(strcmp("dt282x",it->name))
		return 0;

	printk("dac%d: dt282x: 0x%04x ",dev->minor,it->iobase);
	if(check_region(it->iobase,DT2821_SIZE)<0){
		printk("I/O port conflict\n");
		return -EIO;
	}
	request_region(it->iobase,DT2821_SIZE,"dt282x");
	dev->iobase=it->iobase;
	dev->iosize=DT2821_SIZE;

	outw(DT2821_BDINIT,dev->iobase+DT2821_SUPCSR);
	i=inw(dev->iobase+DT2821_ADCSR);
printk("%x,%x,%x,%x,%x ",
		inw(dev->iobase+DT2821_ADCSR),
		inw(dev->iobase+DT2821_CHANCSR),
		inw(dev->iobase+DT2821_DACSR),
		inw(dev->iobase+DT2821_SUPCSR),
		inw(dev->iobase+DT2821_TMRCTR));

	if(
		((inw(dev->iobase+DT2821_ADCSR)&DT2821_ADCSR_MASK)
			!=DT2821_ADCSR_VAL) ||
		((inw(dev->iobase+DT2821_CHANCSR)&DT2821_CHANCSR_MASK)
			!=DT2821_CHANCSR_VAL) ||
		((inw(dev->iobase+DT2821_DACSR)&DT2821_DACSR_MASK)
			!=DT2821_DACSR_VAL) ||
		((inw(dev->iobase+DT2821_SUPCSR)&DT2821_SUPCSR_MASK)
			!=DT2821_SUPCSR_VAL) ||
		((inw(dev->iobase+DT2821_TMRCTR)&DT2821_TMRCTR_MASK)
			!=DT2821_TMRCTR_VAL)){
		printk("not found\n");
		release_region(it->iobase,DT2821_SIZE);
		return -EIO;
	}

	/* should do board test */
	
	irq=it->irq;
	if(irq<0){
		save_flags(flags);
		sti();
		irqs=probe_irq_on();
	
		/* trigger interrupt */

		udelay(100);

		irq=probe_irq_off(irqs);
		restore_flags(flags);
		if(0 /* error */){
			printk("error probing irq (bad) \n");
		}

	}
	dev->irq=0;
	if(irq>0){
		printk("( irq = %d )\n",irq);
		/* request_irq(irq,dt2821_interrupt,SA_INTERRUPT,"dt282x",dev); */
		dev->irq=irq;
	}else if(irq==0){
		printk("(no irq)\n");
	}else{
		printk("(probe returned multiple irqs--bad)\n");
	}
	
	dev->chan=18;

	dev->chinfo=kmalloc(dev->chan*sizeof(dac_chinfo),GFP_KERNEL);
	if(!dev->chinfo){
		return -ENOMEM;
	}
	for(i=0;i<16;i++){
		dev->chinfo[i].flags=DAC_ANALOG|DAC_INPUT;
		dev->chinfo[i].bits=12;
		dev->chinfo[i].trig=dt282x_ad;
		dev->chinfo[i].zero=0.0;
		dev->chinfo[i].delta=10.0/4096;
	}
	dev->chinfo[16].flags=DAC_ANALOG|DAC_OUTPUT;
	dev->chinfo[16].bits=12;
	dev->chinfo[16].trig=dt282x_da;
	dev->chinfo[16].zero=0.0;
	dev->chinfo[16].delta=10.0/4096;

	dev->chinfo[17].flags=DAC_ANALOG|DAC_OUTPUT;
	dev->chinfo[17].bits=12;
	dev->chinfo[17].trig=dt282x_da;
	dev->chinfo[17].zero=0.0;
	dev->chinfo[17].delta=10.0/4096;

	dac_initbuf(&dev->rdbuf);

	dev->rem=dt282x_rem;

	return 1;
}


int dt282x_rem(dac_device *dev)
{
	printk("dac%d: dt282x: remove\n",dev->minor);
	
	kfree(dev->chinfo);

	if(dev->irq){
		free_irq(dev->irq,dev);
	}
	release_region(dev->iobase,dev->iosize);

	dac_freebuf(&dev->rdbuf);
	
	return 0;
}


