/*
 * Driver for Data Translation DT2814, 16 channel D/A converter
 * David A. Schleef
 */

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fcntl.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include "dt2814.h"

static struct dt2814_struct dt2814;
static struct dt2814_channel_struct dt2814_channels[DT2814_NO];
static struct wait_queue *dt2814_wait;

static void dt2814_data(unsigned long minor)
{
	int i,hi,lo;

	outb_p(minor,DT2814_CONTROL);
	while((i=inb_p(DT2814_STATUS))&0x20){
		udelay(25);
	}
	hi=inb_p(DT2814_DATA);
	lo=inb_p(DT2814_DATA);
	DATA(minor)=(hi<<4)+(lo>>4);

	if(FLAGS(minor)&DT2814_VALID){
		FLAGS(minor)|=DT2814_OVRFLW;
	}
	FLAGS(minor)|=DT2814_VALID;
	
	/* assume timer structure doesn't get dorked with */
	TIMER(minor).expires+=DT2814_RATE;
	add_timer(&TIMER(minor));
	wake_up_interruptible(&dt2814_wait);
}

static int dt2814_read(struct inode * inode, struct file * file,char *buf,int nbytes)
{
	struct wait_queue wait={current,NULL};
	unsigned int minor=MINOR(inode->i_rdev);
	unsigned char b[2];
	int n=0;
	int retval=0;

	if(nbytes==0)return 0;

	add_wait_queue(&dt2814_wait,&wait);

	while(nbytes>0){
		current->state=TASK_INTERRUPTIBLE;
		
		if(!(FLAGS(minor)&DT2814_VALID)){
			if(file->f_flags & O_NONBLOCK){
				retval=-EAGAIN;
				break;
			}
			if(current->signal & ~current->blocked){
				retval=-ERESTARTSYS;
				break;
			}
			schedule();
			continue;
		}else{
			if(FLAGS(minor)&DT2814_HIBYTE){
/*printk("dt2814: hibyte (%d)\n",DATA(minor));*/
				b[0]=((DATA(minor)>>8)&0x0f)|(FLAGS(minor)&0xf0);
				memcpy_tofs(buf+n,b,1);
				FLAGS(minor)&=~DT2814_HIBYTE;
				FLAGS(minor)&=~DT2814_VALID;
				FLAGS(minor)&=~DT2814_OVRFLW;
			}else{
/*printk("dt2814: lobyte (%d)\n",DATA(minor));*/
				b[0]=DATA(minor)&0xff;
				memcpy_tofs(buf+n,b,1);
				FLAGS(minor)|=DT2814_HIBYTE;
			}
			n++;
			nbytes--;
			break;  /* I suppose... */
		}
	}
	current->state=TASK_RUNNING;
	remove_wait_queue(&dt2814_wait,&wait);

	return (n ? n:retval);
}

static int dt2814_select(struct inode * inode, struct file * file,int sel_type,
	select_table *wait)
{
	unsigned int minor=MINOR(inode->i_rdev);

	if(sel_type == SEL_IN){
		if(FLAGS(minor)&DT2814_VALID)
			return 1;
		select_wait(&dt2814_wait,wait);
	}
	return 0;
}

static int dt2814_lseek(struct inode * inode, struct file * file,
		off_t offset, int origin)
{
	return -ESPIPE;
}

static int dt2814_open(struct inode * inode, struct file * file)
{
	unsigned int minor=MINOR(inode->i_rdev);
	
	if(minor >= DT2814_NO)
		return -ENODEV;
	if(FLAGS(minor) & DT2814_BUSY)
		return -EBUSY;
	
	init_timer(&TIMER(minor));
	TIMER(minor).expires=jiffies+DT2814_RATE;
	TIMER(minor).data=minor;
	TIMER(minor).function=dt2814_data;
	add_timer(&TIMER(minor));

	FLAGS(minor)&=~DT2814_HIBYTE;

	dt2814_channels[minor].flags |= DT2814_BUSY;
	return 0;
}

static void dt2814_release(struct inode * inode, struct file * file)
{
	unsigned int minor = MINOR(inode->i_rdev);
	
	del_timer(&dt2814_channels[minor].timer);

	dt2814_channels[minor].flags &= ~DT2814_BUSY;
}

static struct file_operations dt2814_fops = {
	dt2814_lseek,
	dt2814_read,
	NULL, /* dt2814_write */
	NULL, /* dt2814_readdir */
	dt2814_select,
	NULL, /* dt2814_ioctl */
	NULL, /* dt2814_mmap */
	dt2814_open,
	dt2814_release
};

static int dt2814_probe(int offset)
{
	int i;

	for(i=0;i<16;i++)DATA(i)=0;

	dt2814.base=offset;
	printk("dt2814: 0x%04x 0x%08lx\n",offset,dt2814.base);
	if(check_region(offset,2)<0)return -1;
	request_region(offset,2,"dt2814");

	outb_p(0,DT2814_CONTROL);
	udelay(100);
	i=inb_p(DT2814_DATA);
	i=inb_p(DT2814_DATA);

	return 1;
}

int dt2814_init(void)
{
	int ret;

	if(register_chrdev(DT2814_MAJOR,"dt2814",&dt2814_fops)){
		printk("dt2814: unable to get major %d\n",DT2814_MAJOR);
		return -EIO;
	}
	
	ret=dt2814_probe(DT2814_BASE);
	dt2814_wait=NULL;

	return 0;
}

